import logging
import os
import sys
import typing
from logging.handlers import TimedRotatingFileHandler

import matplotlib as mpl
import matplotlib.pyplot as plt
from PySide6 import QtCore, QtGui, QtWidgets
from qt_material import apply_stylesheet, list_themes

from .. import QGRAIN_ROOT_PATH, QGRAIN_VERSION
from ..chart import setup_matplotlib
from ..io import save_pca, save_statistic
from ..model import GrainSizeDataset
from .AboutDialog import AboutDialog
from .ClusteringAnalyzer import ClusteringAnalyzer
from .DatasetGenerator import DatasetGenerator
from .EMMAAnalyzer import EMMAAnalyzer
from .EMMASettingDialog import EMMASettingDialog
from .GrainSizeDatasetViewer import GrainSizeDatasetViewer
from .LoadDatasetDialog import LoadDatasetDialog
from .LogDialog import *
from .ParameterEditor import ParameterEditor
from .PCAAnalyzer import PCAAnalyzer
from .SSUAnalyzer import SSUAnalyzer
from .SSUMulticoreAnalyzer import SSUMulticoreAnalyzer
from .SSUSettingDialog import SSUSettingDialog
from .UDMAnalyzer import UDMAnalyzer
from .UDMSettingDialog import UDMSettingDialog


class MainWindow(QtWidgets.QMainWindow):
    logger = logging.getLogger("QGrain.MainWindow")
    def __init__(self):
        super().__init__()
        self.setWindowTitle("QGrain")
        self.__dataset = GrainSizeDataset() # type: GrainSizeDataset
        self.current_translator = None # type: QtCore.QTranslator
        self.ssu_setting_dialog = SSUSettingDialog(self)
        self.emma_setting_dialog = EMMASettingDialog(self)
        self.udm_setting_dialog = UDMSettingDialog(self)
        self.parameter_editor = ParameterEditor(self)
        self.ssu_multicore_analyzer = SSUMulticoreAnalyzer(self)
        self.init_ui()
        self.ssu_multicore_analyzer.result_finished.connect(self.ssu_analyzer.result_view.add_result)
        self.load_dataset_dialog = LoadDatasetDialog(self)
        self.load_dataset_dialog.dataset_loaded.connect(self.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.dataset_viewer.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.pca_analyzer.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.clustering_analyzer.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.ssu_analyzer.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.emma_analyzer.on_dataset_loaded)
        self.load_dataset_dialog.dataset_loaded.connect(self.udm_analyzer.on_dataset_loaded)
        self.log_dialog = LogDialog(self)
        self.about_dialog = AboutDialog(self)
        self.file_dialog = QtWidgets.QFileDialog(parent=self)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self.close_msg = QtWidgets.QMessageBox(self)
        self.close_msg.setWindowTitle(self.tr("Warning"))
        self.close_msg.setText(self.tr("Closing this window will terminate all running tasks, are you sure to close it?"))
        self.close_msg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        self.close_msg.setDefaultButton(QtWidgets.QMessageBox.No)

        # load the artificial dataset to show the functions of all modules
        dataset = self.dataset_generator.get_random_dataset(100)
        self.load_dataset_dialog.dataset_loaded.emit(dataset.dataset_to_fit)
        self.ssu_analyzer.on_try_fit_clicked()
        self.emma_analyzer.on_try_fit_clicked()
        self.udm_analyzer.on_try_fit_clicked()

    def init_ui(self):
        self.tab_widget = QtWidgets.QTabWidget(self)
        self.tab_widget.setTabPosition(QtWidgets.QTabWidget.West)
        self.setCentralWidget(self.tab_widget)
        self.dataset_generator = DatasetGenerator()
        self.tab_widget.addTab(self.dataset_generator, self.tr("Generator"))
        self.dataset_viewer = GrainSizeDatasetViewer()
        self.tab_widget.addTab(self.dataset_viewer, self.tr("Statistics"))
        self.pca_analyzer = PCAAnalyzer()
        self.tab_widget.addTab(self.pca_analyzer, self.tr("PCA"))
        self.clustering_analyzer = ClusteringAnalyzer()
        self.tab_widget.addTab(self.clustering_analyzer, self.tr("Clustering"))
        self.ssu_analyzer = SSUAnalyzer(self.ssu_setting_dialog, self.parameter_editor)
        self.tab_widget.addTab(self.ssu_analyzer, self.tr("SSU"))
        self.emma_analyzer = EMMAAnalyzer(self.emma_setting_dialog, self.parameter_editor)
        self.tab_widget.addTab(self.emma_analyzer, self.tr("EMMA"))
        self.udm_analyzer = UDMAnalyzer(self.udm_setting_dialog, self.parameter_editor)
        self.tab_widget.addTab(self.udm_analyzer, self.tr("UDM"))
        self.init_menus()

    def init_menus(self):
        # Open
        self.open_menu = self.menuBar().addMenu(self.tr("Open")) # type: QtWidgets.QMenu
        self.open_dataset_action = self.open_menu.addAction(self.tr("Grain Size Dataset")) # type: QtGui.QAction
        self.open_dataset_action.triggered.connect(lambda: self.load_dataset_dialog.show())
        self.load_ssu_result_action = self.open_menu.addAction(self.tr("SSU Results")) # type: QtGui.QAction
        self.load_ssu_result_action.triggered.connect(self.ssu_analyzer.result_view.load_results)
        self.load_ssu_reference_action = self.open_menu.addAction(self.tr("SSU References")) # type: QtGui.QAction
        self.load_ssu_reference_action.triggered.connect(self.ssu_analyzer.reference_view.load_references)
        self.load_emma_result_action = self.open_menu.addAction(self.tr("EMMA Result")) # type: QtGui.QAction
        self.load_emma_result_action.triggered.connect(self.emma_analyzer.load_result)
        self.load_udm_result_action = self.open_menu.addAction(self.tr("UDM Result")) # type: QtGui.QAction
        self.load_udm_result_action.triggered.connect(self.udm_analyzer.load_result)

        # Save
        self.save_menu = self.menuBar().addMenu(self.tr("Save")) # type: QtWidgets.QMenu
        self.save_statistic_action = self.save_menu.addAction(self.tr("Statistic Result")) # type: QtGui.QAction
        self.save_statistic_action.triggered.connect(self.on_save_statistic_clicked)
        self.save_pca_action = self.save_menu.addAction(self.tr("PCA Result")) # type: QtGui.QAction
        self.save_pca_action.triggered.connect(self.on_save_pca_clicked)
        self.save_clustering_action = self.save_menu.addAction(self.tr("Clustering Result")) # type: QtGui.QAction
        self.save_clustering_action.triggered.connect(self.clustering_analyzer.save_result)
        self.save_ssu_result_action = self.save_menu.addAction(self.tr("SSU Results")) # type: QtGui.QAction
        self.save_ssu_result_action.triggered.connect(self.ssu_analyzer.result_view.save_results)
        self.save_ssu_reference_action = self.save_menu.addAction(self.tr("SSU References")) # type: QtGui.QAction
        self.save_ssu_reference_action.triggered.connect(self.ssu_analyzer.reference_view.dump_references)
        self.save_emma_result_action = self.save_menu.addAction(self.tr("EMMA Result")) # type: QtGui.QAction
        self.save_emma_result_action.triggered.connect(self.emma_analyzer.save_selected_result)
        self.save_udm_result_action = self.save_menu.addAction(self.tr("UDM Result")) # type: QtGui.QAction
        self.save_udm_result_action.triggered.connect(self.udm_analyzer.save_selected_result)

        # Config
        self.config_menu = self.menuBar().addMenu(self.tr("Config")) # type: QtWidgets.QMenu
        self.config_ssu_action = self.config_menu.addAction(self.tr("SSU Settings")) # type: QtGui.QAction
        self.config_ssu_action.triggered.connect(self.ssu_setting_dialog.show)
        self.config_emma_action = self.config_menu.addAction(self.tr("EMMA Settings")) # type: QtGui.QAction
        self.config_emma_action.triggered.connect(self.emma_setting_dialog.show)
        self.config_udm_action = self.config_menu.addAction(self.tr("UDM Settings")) # type: QtGui.QAction
        self.config_udm_action.triggered.connect(self.udm_setting_dialog.show)

        # Experimental
        self.experimental_menu = self.menuBar().addMenu(self.tr("Experimental")) # type: QtWidgets.QMenu
        self.ssu_fit_all_action = self.experimental_menu.addAction(self.tr("Perform SSU For All Samples")) # type: QtGui.QAction
        self.ssu_fit_all_action.triggered.connect(self.ssu_fit_all_samples)

        # Language
        self.language_menu = self.menuBar().addMenu(self.tr("Language")) # type: QtWidgets.QMenu
        self.language_group = QtGui.QActionGroup(self.language_menu)
        self.language_group.setExclusive(True)
        self.language_actions = [] # type: list[QtGui.QAction]
        for key, name in self.supported_languages:
            action = self.language_group.addAction(name)
            action.setCheckable(True)
            action.triggered.connect(lambda checked=False, language=key: self.switch_language(language))
            self.language_menu.addAction(action)
            self.language_actions.append(action)
        self.language_actions[0].setChecked(True)
        # Theme
        self.theme_menu = self.menuBar().addMenu(self.tr("Theme")) # type: QtWidgets.QMenu
        self.theme_group = QtGui.QActionGroup(self.theme_menu)
        self.theme_group.setExclusive(True)
        self.theme_actions = []
        for theme in list_themes():
            action = self.theme_group.addAction(theme)
            action.setCheckable(True)
            app = QtCore.QCoreApplication.instance()
            invert_secondary = theme.startswith("light")
            action.triggered.connect(lambda checked=False, theme=theme, invert_secondary=invert_secondary: apply_stylesheet(app, theme=theme, invert_secondary=invert_secondary, extra=EXTRA))
            self.theme_menu.addAction(action)
            self.theme_actions.append(action)
            if theme == "dark_cyan.xml":
                action.setChecked(True)
        # Log
        self.log_action = QtGui.QAction(self.tr("Log"))
        self.log_action.triggered.connect(lambda: self.log_dialog.show())
        self.menuBar().addAction(self.log_action)
        # About
        self.about_action = QtGui.QAction(self.tr("About"))
        self.about_action.triggered.connect(lambda: self.about_dialog.show())
        self.menuBar().addAction(self.about_action)

    @property
    def supported_languages(self) -> typing.List[typing.Tuple[str, str]]:
        languages = [("en", "English"),
                     ("zh_CN", "简体中文")]
        return languages

    @property
    def language(self) -> str:
        for i, language_action in enumerate(self.language_actions):
            if language_action.isChecked():
                key, name = self.supported_languages[i]
                return key

    @property
    def theme(self) -> str:
        for i, theme_action in enumerate(self.theme_actions):
            if theme_action.isChecked():
                theme = list_themes()[i]
                return theme

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def on_dataset_loaded(self, dataset: GrainSizeDataset):
        if dataset is None or not dataset.has_sample:
            return
        self.__dataset = dataset

    def ssu_fit_all_samples(self):
        tasks = self.ssu_analyzer.get_all_tasks()
        self.ssu_multicore_analyzer.setup_tasks(tasks)
        self.ssu_multicore_analyzer.exec()

    def on_save_statistic_clicked(self):
        if not self.__dataset.has_sample:
            self.logger.error("Dataset has not been loaded.")
            self.show_error(self.tr("Dataset has not been loaded."))
            return
        filename, _ = self.file_dialog.getSaveFileName(
            self, self.tr("Save statistic result"),
            None, "Microsoft Excel (*.xlsx)")
        if filename is None or filename == "":
            return
        try:
            progress_dialog = QtWidgets.QProgressDialog(
                self.tr("Saving statistic result..."), self.tr("Cancel"),
                0, 100, self)
            progress_dialog.setWindowTitle("QGrain")
            progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
            def callback(progress: float):
                if progress_dialog.wasCanceled():
                    raise StopIteration()
                progress_dialog.setValue(int(progress*100))
                QtCore.QCoreApplication.processEvents()
            save_statistic(self.__dataset, filename, progress_callback=callback, logger=self.logger)
        except Exception as e:
            self.logger.exception("An unknown exception was raised. Please check the logs for more details.", stack_info=True)
            self.show_error(self.tr("An unknown exception was raised. Please check the logs for more details."))

    def on_save_pca_clicked(self):
        if not self.__dataset.has_sample:
            self.show_warning(self.tr("Dataset has not been loaded."))
            return
        filename, _ = self.file_dialog.getSaveFileName(
            None, self.tr("Save PCA result"),
            None, "Microsoft Excel (*.xlsx)")
        if filename is None or filename == "":
            return
        try:
            progress_dialog = QtWidgets.QProgressDialog(
                self.tr("Saving PCA result..."), self.tr("Cancel"),
                0, 100, self)
            progress_dialog.setWindowTitle("QGrain")
            progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
            def callback(progress: float):
                if progress_dialog.wasCanceled():
                    raise StopIteration()
                progress_dialog.setValue(int(progress*100))
                QtCore.QCoreApplication.processEvents()
            save_pca(self.__dataset, filename, progress_callback=callback, logger=self.logger)
        except Exception as e:
            self.logger.exception("An unknown exception was raised. Please check the logs for more details.", stack_info=True)
            self.show_error(self.tr("An unknown exception was raised. Please check the logs for more details."))

    def switch_language(self, language: str):
        app = QtWidgets.QApplication.instance()
        if self.current_translator is not None:
            app.removeTranslator(self.current_translator)
        translator = QtCore.QTranslator(app)
        translator.load(os.path.join(QGRAIN_ROOT_PATH, "assets", language))
        app.installTranslator(translator)
        self.current_translator = translator

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def closeEvent(self, event: QtGui.QCloseEvent):
        res = self.close_msg.exec_()
        if res == QtWidgets.QMessageBox.Yes:
            event.accept()
        else:
            event.ignore()

    def retranslate(self):
        self.setWindowTitle("QGrain")
        self.close_msg.setWindowTitle(self.tr("Warning"))
        self.close_msg.setText(self.tr("Closing this window will terminate all running tasks, are you sure to close it?"))
        self.open_menu.setTitle((self.tr("Open")))
        self.save_menu.setTitle(self.tr("Save"))
        self.open_dataset_action.setText(self.tr("Grain Size Dataset"))
        self.load_ssu_result_action.setText(self.tr("SSU Results"))
        self.load_ssu_reference_action.setText(self.tr("SSU References"))
        self.load_emma_result_action.setText(self.tr("EMMA Result"))
        self.load_udm_result_action.setText(self.tr("UDM Result"))
        self.save_statistic_action.setText(self.tr("Statistic Result"))
        self.save_pca_action.setText(self.tr("PCA Result"))
        self.save_clustering_action.setText(self.tr("Clustering Result"))
        self.save_ssu_result_action.setText(self.tr("SSU Results"))
        self.save_ssu_reference_action.setText(self.tr("SSU References"))
        self.save_emma_result_action.setText(self.tr("EMMA Result"))
        self.save_udm_result_action.setText(self.tr("UDM Result"))
        self.config_menu.setTitle(self.tr("Config"))
        self.config_ssu_action.setText(self.tr("SSU Settings"))
        self.config_emma_action.setText(self.tr("EMMA Settings"))
        self.config_udm_action.setText(self.tr("UDM Settings"))
        self.experimental_menu.setTitle(self.tr("Experimental"))
        self.ssu_fit_all_action.setText(self.tr("Perform SSU For All Samples"))
        self.language_menu.setTitle(self.tr("Language"))
        self.theme_menu.setTitle(self.tr("Theme"))
        self.log_action.setText(self.tr("Log"))
        self.about_action.setText(self.tr("About"))
        self.tab_widget.setTabText(0, self.tr("Generator"))
        self.tab_widget.setTabText(1, self.tr("Statistics"))
        self.tab_widget.setTabText(2, self.tr("PCA"))
        self.tab_widget.setTabText(3, self.tr("Clustering"))
        self.tab_widget.setTabText(4, self.tr("SSU"))
        self.tab_widget.setTabText(5, self.tr("EMMA"))
        self.tab_widget.setTabText(6, self.tr("UDM"))
        self.ssu_multicore_analyzer.retranslate()


EXTRA = {'font_family': 'Roboto,Arial,Helvetica,Tahoma,Verdana,Microsoft YaHei UI,SimSum'}

def getdirsize(dir):
    size = 0
    for root, _, files in os.walk(dir):
        size += sum([os.path.getsize(os.path.join(root, name)) for name in files])
    return size

def create_necessary_folders():
    necessary_folders = (
        os.path.join(os.path.expanduser("~"), "QGrain"),
        os.path.join(os.path.expanduser("~"), "QGrain", "logs"))
    for folder in necessary_folders:
        os.makedirs(folder, exist_ok=True)

def setup_language(app: QtWidgets.QApplication, language: str):
    trans = QtCore.QTranslator(app)
    trans.load(os.path.join(QGRAIN_ROOT_PATH, "assets", language))
    app.installTranslator(trans)

def setup_logging(main: MainWindow):
    format_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    file_handler = TimedRotatingFileHandler(os.path.join(os.path.expanduser("~"), "QGrain", "logs", "qgrain.log"), when="D", backupCount=8, encoding="utf-8")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(logging.Formatter(format_str))
    status_bar_handler = StatusBarLogHandler(main.statusBar(), level=logging.INFO)
    status_bar_handler.setFormatter(logging.Formatter(format_str))
    gui_handler = GUILogHandler(main.log_dialog, level=logging.DEBUG)
    gui_handler.setFormatter(logging.Formatter(format_str))
    logging.basicConfig(level=logging.DEBUG, format=format_str)
    logging.getLogger().addHandler(file_handler)
    logging.getLogger().addHandler(status_bar_handler)
    logging.getLogger().addHandler(gui_handler)
    mpl.set_loglevel("error")

def setup_app():
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_DisableHighDpiScaling)
    app = QtWidgets.QApplication(sys.argv)
    # for filename in os.listdir(os.path.join(QGRAIN_ROOT_PATH, "assets")):
    #     if filename.endswith(".ttf"):
    #         font_path = os.path.join(QGRAIN_ROOT_PATH, "assets", filename)
    #         font_id = QtGui.QFontDatabase.addApplicationFont(font_path)
    #         font_family = QtGui.QFontDatabase.applicationFontFamilies(font_id)
    splash = QtWidgets.QSplashScreen()
    pixmap = QtGui.QPixmap(os.path.join(QGRAIN_ROOT_PATH, "assets", "icon.png"))
    pixmap.setDevicePixelRatio(1.0)
    splash.setPixmap(pixmap)
    splash.show()

    create_necessary_folders()
    app.setWindowIcon(QtGui.QIcon(pixmap))
    # app.setApplicationDisplayName(f"QGrain ({QGRAIN_VERSION})")
    app.setApplicationVersion(QGRAIN_VERSION)
    from qt_material import apply_stylesheet
    apply_stylesheet(app, theme='dark_cyan.xml', invert_secondary=False, extra=EXTRA)
    setup_matplotlib()
    setup_language(app, "en")
    return app, splash

def qgrain_app():
    app, splash = setup_app()
    main = MainWindow()
    setup_logging(main)
    # main.move(main.screen().geometry().center() - main.frameGeometry().center())
    main.show()
    splash.finish(main)
    sys.exit(app.exec_())



from __future__ import annotations

import copy
import typing

import numpy as np

from ..emma import KernelType
from ..model import GrainSizeDataset
from ..ssu import (DISTRIBUTION_CLASS_MAP, DistributionType, GeneralWeibull,
                   Normal, SkewNormal, Weibull, get_distance_function)
from ._setting import UDMAlgorithmSetting


class UDMResult:
    def __init__(self,
                 dataset: GrainSizeDataset,
                 kernel_type: KernelType,
                 n_components: int,
                 initial_parameters: np.ndarray,
                 resolver_setting: UDMAlgorithmSetting,
                 distribution_loss_series: typing.Iterable[float],
                 component_loss_series: typing.Iterable[float],
                 time_spent: float,
                 final_parameters: np.ndarray,
                 history: typing.Iterable[np.ndarray]):
        self.__dataset = dataset
        self.__kernel_type = kernel_type
        self.__n_components = n_components
        self.__initial_parameters = initial_parameters
        self.__resolver_setting = resolver_setting
        self.__distribution_loss_series = distribution_loss_series
        self.__component_loss_series = component_loss_series
        self.__time_spent = time_spent
        self.__final_parameters = final_parameters
        self.__history = [final_parameters] if history is None else history
        self.__classes = np.expand_dims(np.expand_dims(self.dataset.classes_φ, axis=0), axis=0).repeat(self.n_samples, axis=0).repeat(self.n_components, axis=1)
        self.__interval = np.abs((self.dataset.classes_φ[0]-self.dataset.classes_φ[-1]) / (self.n_classes-1))
        self.__distribution_class = DISTRIBUTION_CLASS_MAP[DistributionType.__members__[self.kernel_type.name]]
        self.update(final_parameters)

    @property
    def dataset(self) -> GrainSizeDataset:
        return self.__dataset

    @property
    def n_samples(self) -> int:
        return self.dataset.n_samples

    @property
    def n_classes(self) -> int:
        return len(self.dataset.classes_μm)

    @property
    def kernel_type(self) -> KernelType:
        return self.__kernel_type

    @property
    def n_components(self) -> int:
        return self.__n_components

    @property
    def initial_parameters(self) -> np.ndarray:
        return self.__initial_parameters

    @property
    def resolver_setting(self) -> UDMAlgorithmSetting:
        return self.__resolver_setting

    @property
    def proportions(self) -> np.ndarray:
        return self.__proportions

    @property
    def components(self) -> np.ndarray:
        return self.__components

    @property
    def distribution_loss_series(self) -> np.ndarray:
        return self.__distribution_loss_series

    @property
    def component_loss_series(self) -> np.ndarray:
        return self.__component_loss_series

    @property
    def final_parameters(self) -> np.ndarray:
        return self.__final_parameters

    @property
    def time_spent(self) -> float:
        return self.__time_spent

    @property
    def n_iterations(self) -> int:
        return len(self.__history)

    @property
    def history(self) -> typing.Iterable[UDMResult]:
        for parameters in self.__history:
            copy_result = copy.copy(self)
            self.update(parameters)
            yield copy_result

    def update(self, parameters: np.ndarray):
        proportions, components, mvsk = self.__distribution_class.interpret(parameters, self.__classes, self.__interval)
        self.__proportions = proportions
        self.__components = components

    def get_distance(self, distance: str) -> float:
        distance_func = get_distance_function(distance)
        predict = (self.proportions @ self.components).squeeze()
        return distance_func(predict, self.dataset.distribution_matrix)



from enum import Enum, unique

import numpy as np
import torch

_INFINITESIMAL = 1e-8


@unique
class KernelType(Enum):
    Nonparametric = "Nonparametric"
    Normal = "Normal"
    SkewNormal = "Skew Normal"
    Weibull = "Weibull"
    GeneralWeibull = "General Weibull"


def normal_pdf(x, loc, scale):
    pdf = 1 / (scale*np.sqrt(2*np.pi)) * torch.exp(-torch.square(x - loc) / (2*scale**2))
    return pdf


def std_normal_cdf(x):
    cdf = 0.5 * (1 + torch.erf(x / np.sqrt(2)))
    return cdf


def skew_normal_pdf(x, shape, loc, scale):
    pdf = 2 * normal_pdf(x, loc, scale) * std_normal_cdf(shape*(x-loc)/scale)
    return pdf


def weibull_pdf(x, shape, scale):
    key = torch.greater_equal(x, 0)
    pdf = torch.zeros_like(x)
    pdf[key] = (shape[key] / scale[key]) * (x[key] / scale[key]) ** (shape[key]-1) * torch.exp(-(x[key]/scale[key])**shape[key])
    return pdf


def general_weibull_pdf(x, shape, loc, scale):
    y = x - loc
    key = torch.greater_equal(y, 0)
    pdf = torch.zeros_like(y)
    pdf[key] = (shape[key] / scale[key]) * (y[key] / scale[key]) ** (shape[key]-1) * torch.exp(-(y[key]/scale[key])**shape[key])
    return pdf


class NonparametricKernel(torch.nn.Module):
    def __init__(self, n_samples: int, n_members: int, n_classes: int, params: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = n_classes
        if params is None:
            self.params = torch.nn.Parameter(torch.randn(n_samples, n_members, n_classes)+2.0, requires_grad=True)
        else:
            assert params.ndim == 2
            assert params.shape == (n_members, n_classes)
            self.params = torch.nn.Parameter(torch.from_numpy(np.expand_dims(params, 0).repeat(n_samples, 0)), requires_grad=True)

    def forward(self, classes: torch.Tensor, interval: float):
        assert classes.shape == (self.n_samples, self.n_members, self.n_classes)
        end_members = torch.softmax(self.params, dim=-1)
        return end_members

class NormalKernel(torch.nn.Module):
    N_PARAMS = 2
    def __init__(self, n_samples: int, n_members: int, n_classes: int, params: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = n_classes
        if params is None:
            self.params = torch.nn.Parameter(torch.randn(n_samples, self.N_PARAMS, n_members)+5.0, requires_grad=True)
        else:
            assert params.ndim == 2
            assert params.shape == (self.N_PARAMS, n_members)
            self.params = torch.nn.Parameter(torch.from_numpy(np.expand_dims(params, 0).repeat(n_samples, 0)), requires_grad=True)

    def forward(self, classes: torch.Tensor, interval: float):
        assert classes.shape == (self.n_samples, self.n_members, self.n_classes)
        locations = self.params[:, 0, :].unsqueeze(2).repeat(1, 1, self.n_classes)
        scales = (torch.relu(self.params[:, 1, :])+_INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        end_members = normal_pdf(classes, locations, scales) * interval
        # params = torch.cat([locations[:, :, 0].unsqueeze(1), scales[:, :, 0].unsqueeze(1)], dim=1)
        return end_members


class SkewNormalKernel(torch.nn.Module):
    N_PARAMS = 3
    def __init__(self, n_samples: int, n_members: int, n_classes: int, params: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = n_classes
        if params is None:
            self.params = torch.nn.Parameter(torch.randn(n_samples, self.N_PARAMS, n_members)+5.0, requires_grad=True)
        else:
            assert params.ndim == 2
            assert params.shape == (self.N_PARAMS, n_members)
            self.params = torch.nn.Parameter(torch.from_numpy(np.expand_dims(params, 0).repeat(n_samples, 0)), requires_grad=True)

    def forward(self, classes: torch.Tensor, interval: float):
        assert classes.shape == (self.n_samples, self.n_members, self.n_classes)
        shapes = self.params[:, 0, :].unsqueeze(2).repeat(1, 1, self.n_classes)
        locations = self.params[:, 1, :].unsqueeze(2).repeat(1, 1, self.n_classes)
        scales = (torch.relu(self.params[:, 2, :])+_INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        end_members = skew_normal_pdf(classes, shapes, locations, scales) * interval
        # params = torch.cat([shapes[:, :, 0].unsqueeze(1), locations[:, :, 0].unsqueeze(1), scales[:, :, 0].unsqueeze(1)], dim=1)
        return end_members


class WeibullKernel(torch.nn.Module):
    N_PARAMS = 2
    def __init__(self, n_samples: int, n_members: int, n_classes: int, params: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = n_classes
        if params is None:
            self.params = torch.nn.Parameter(torch.randn(n_samples, self.N_PARAMS, n_members)+5.0, requires_grad=True)
        else:
            assert params.ndim == 2
            assert params.shape == (self.N_PARAMS, n_members)
            self.params = torch.nn.Parameter(torch.from_numpy(np.expand_dims(params, 0).repeat(n_samples, 0)), requires_grad=True)

    def forward(self, classes: torch.Tensor, interval: float):
        assert classes.shape == (self.n_samples, self.n_members, self.n_classes)
        shapes = (torch.relu(self.params[:, 0, :])+_INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        scales = (torch.relu(self.params[:, 1, :])+_INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        end_members = weibull_pdf(classes, shapes, scales) * interval
        # params = torch.cat([shapes[:, :, 0].unsqueeze(1), scales[:, :, 0].unsqueeze(1)], dim=1)
        return end_members


class GeneralWeibullKernel(torch.nn.Module):
    N_PARAMS = 3
    def __init__(self, n_samples: int, n_members: int, n_classes: int, params: np.ndarray = None):
        super().__init__()
        self.n_samples = n_samples
        self.n_members = n_members
        self.n_classes = n_classes

        if params is None:
            self.params = torch.nn.Parameter(torch.randn(n_samples, self.N_PARAMS, n_members)+5.0, requires_grad=True)
        else:
            assert params.ndim == 2
            assert params.shape == (self.N_PARAMS, n_members)
            self.params = torch.nn.Parameter(torch.from_numpy(np.expand_dims(params, 0).repeat(n_samples, 0)), requires_grad=True)

    def forward(self, classes: torch.Tensor, interval: float):
        assert classes.shape == (self.n_samples, self.n_members, self.n_classes)
        shapes = (torch.relu(self.params[:, 0, :])+_INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        locations = self.params[:, 1, :].unsqueeze(2).repeat(1, 1, self.n_classes)
        scales = (torch.relu(self.params[:, 2, :])+_INFINITESIMAL).unsqueeze(2).repeat(1, 1, self.n_classes)
        end_members = general_weibull_pdf(classes, shapes, locations, scales) * interval
        # params = torch.cat([shapes[:, :, 0].unsqueeze(1), locations[:, :, 0].unsqueeze(1), scales[:, :, 0].unsqueeze(1)], dim=1)
        return end_members

class Proportion(torch.nn.Module):
    def __init__(self, n_samples: int, n_members: int):
        super().__init__()
        self.params = torch.nn.Parameter(torch.rand(n_samples, 1, n_members), requires_grad=True)

    def forward(self):
        # n_samples x 1 x n_members
        proportions = torch.softmax(self.params, 2)
        return proportions


KERNEL_CLASS_MAP = {
    KernelType.Nonparametric: NonparametricKernel,
    KernelType.Normal: NormalKernel,
    KernelType.SkewNormal: SkewNormalKernel,
    KernelType.Weibull: WeibullKernel,
    KernelType.GeneralWeibull: GeneralWeibullKernel}



__all__ = ["SSUReferenceViewer"]

import logging
import pickle
import time
import typing
from uuid import UUID

import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets

from ..chart.DistanceCurveChart import DistanceCurveChart
from ..model import GrainSizeSample
from ..ssu import SSUResult, built_in_distances, get_distance_function


class SSUReferenceViewer(QtWidgets.QWidget):
    PAGE_ROWS = 20
    logger = logging.getLogger("QGrain.SSUReferenceViewer")
    result_displayed = QtCore.Signal(SSUResult, bool)
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.__fitting_results = [] # type: list[SSUResult]
        self.__reference_map = {} # type: dict[UUID, SSUResult]
        self.init_ui()
        self.distance_chart = DistanceCurveChart()
        self.file_dialog = QtWidgets.QFileDialog(parent=self)
        self.update_page_list()
        self.update_page(self.page_index)
        self.remove_warning_msg = QtWidgets.QMessageBox(self)
        self.remove_warning_msg.setStandardButtons(QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Yes)
        self.remove_warning_msg.setDefaultButton(QtWidgets.QMessageBox.No)
        self.remove_warning_msg.setWindowTitle(self.tr("Warning"))
        self.remove_warning_msg.setText(self.tr("Are you sure to remove all SSU results?"))
        self.normal_msg = QtWidgets.QMessageBox(self)

    def init_ui(self):
        self.setWindowTitle(self.tr("SSU Reference Viewer"))
        self.data_table = QtWidgets.QTableWidget(100, 100)
        self.data_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.data_table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.data_table.setAlternatingRowColors(True)
        self.data_table.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.main_layout.addWidget(self.data_table, 0, 0, 1, 3)

        self.previous_button = QtWidgets.QPushButton(self.tr("Previous"))
        self.previous_button.setToolTip(self.tr("Click to get back to the previous page."))
        self.page_combo_box = QtWidgets.QComboBox()
        self.page_combo_box.addItem(self.tr("Page {0}").format(1))
        self.next_button = QtWidgets.QPushButton(self.tr("Next"))
        self.next_button.setToolTip(self.tr("Click to jump to the next page."))
        self.previous_button.clicked.connect(lambda: self.page_combo_box.setCurrentIndex(max(self.page_index-1, 0)))
        self.page_combo_box.currentIndexChanged.connect(self.update_page)
        self.next_button.clicked.connect(lambda: self.page_combo_box.setCurrentIndex(min(self.page_index+1, self.n_pages-1)))
        self.main_layout.addWidget(self.previous_button, 1, 0)
        self.main_layout.addWidget(self.page_combo_box, 1, 1)
        self.main_layout.addWidget(self.next_button, 1, 2)

        self.distance_label = QtWidgets.QLabel(self.tr("Distance Function"))
        self.distance_label.setToolTip(self.tr("The function to calculate the difference (on the contrary, similarity) between two samples."))
        self.distance_combo_box = QtWidgets.QComboBox()
        self.distance_combo_box.addItems(built_in_distances)
        self.distance_combo_box.setCurrentText("log10MSE")
        self.distance_combo_box.currentTextChanged.connect(lambda: self.update_page(self.page_index))
        self.main_layout.addWidget(self.distance_label, 2, 0)
        self.main_layout.addWidget(self.distance_combo_box, 2, 1, 1, 2)
        self.menu = QtWidgets.QMenu(self.data_table)
        self.mark_action = self.menu.addAction(self.tr("Mark")) # type: QtGui.QAction
        self.mark_action.triggered.connect(self.mark_selections)
        self.unmark_action = self.menu.addAction(self.tr("Unmark")) # type: QtGui.QAction
        self.unmark_action.triggered.connect(self.unmark_selections)
        self.remove_action = self.menu.addAction(self.tr("Remove")) # type: QtGui.QAction
        self.remove_action.triggered.connect(self.remove_selections)
        self.remove_all_action = self.menu.addAction(self.tr("Remove All")) # type: QtGui.QAction
        self.remove_all_action.triggered.connect(self.remove_all_results)
        self.show_chart_action = self.menu.addAction(self.tr("Show Chart")) # type: QtGui.QAction
        self.show_chart_action.triggered.connect(self.show_chart)
        self.show_distance_action = self.menu.addAction(self.tr("Show Distance Series")) # type: QtGui.QAction
        self.show_distance_action.triggered.connect(self.show_distance)
        self.data_table.customContextMenuRequested.connect(self.show_menu)
        # necessary to add actions of menu to this widget itself,
        # otherwise, the shortcuts will not be triggered
        self.addActions(self.menu.actions())

    def show_menu(self, pos: QtCore.QPoint):
        self.menu.popup(QtGui.QCursor.pos())

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    @property
    def distance_name(self) -> str:
        return self.distance_combo_box.currentText()

    @property
    def distance_function(self) -> typing.Callable:
        return get_distance_function(self.distance_combo_box.currentText())

    @property
    def page_index(self) -> int:
        return self.page_combo_box.currentIndex()

    @property
    def n_pages(self) -> int:
        return self.page_combo_box.count()

    @property
    def n_results(self) -> int:
        return len(self.__fitting_results)

    @property
    def selections(self) -> typing.List[int]:
        start = self.page_index*self.PAGE_ROWS
        temp = set()
        for item in self.data_table.selectedRanges():
            for i in range(item.topRow(), min(self.PAGE_ROWS+1, item.bottomRow()+1)):
                temp.add(i+start)
        indexes = list(temp)
        indexes.sort()
        return indexes

    def update_page_list(self):
        last_page_index = self.page_index
        if self.n_results == 0:
            n_pages = 1
        else:
            n_pages, left = divmod(self.n_results, self.PAGE_ROWS)
            if left != 0:
                n_pages += 1
        self.page_combo_box.blockSignals(True)
        self.page_combo_box.clear()
        self.page_combo_box.addItems([self.tr("Page {0}").format(i+1) for i in range(n_pages)])
        if last_page_index >= n_pages:
            self.page_combo_box.setCurrentIndex(n_pages-1)
        else:
            self.page_combo_box.setCurrentIndex(last_page_index)
        self.page_combo_box.blockSignals(False)

    def update_page(self, page_index: int):
        def write(row: int, col: int, value: str):
            if isinstance(value, str):
                pass
            elif isinstance(value, int):
                value = str(value)
            elif isinstance(value, float):
                value = f"{value: 0.4f}"
            else:
                value = value.__str__()
            item = QtWidgets.QTableWidgetItem(value)
            item.setTextAlignment(QtCore.Qt.AlignCenter)
            self.data_table.setItem(row, col, item)
        # necessary to clear
        self.data_table.clear()
        if page_index == self.n_pages - 1:
            start = page_index * self.PAGE_ROWS
            end = self.n_results
        else:
            start, end = page_index * self.PAGE_ROWS, (page_index+1) * self.PAGE_ROWS
        self.data_table.setRowCount(end-start)
        self.data_table.setColumnCount(7)
        self.data_table.setHorizontalHeaderLabels([
            self.tr("Distribution Type"),
            self.tr("Number of Components"),
            self.tr("Number of Iterations"),
            self.tr("Spent Time [s]"),
            self.tr("Final Distance"),
            self.tr("Has Reference"),
            self.tr("Is Reference")])
        self.data_table.horizontalHeader().setDefaultAlignment(QtCore.Qt.AlignCenter | QtCore.Qt.TextWordWrap)
        sample_names = [result.sample.name for result in self.__fitting_results[start: end]]
        self.data_table.setVerticalHeaderLabels(sample_names)
        for row, result in enumerate(self.__fitting_results[start: end]):
            write(row, 0, result.task.distribution_type.value)
            write(row, 1, result.task.n_components)
            write(row, 2, result.n_iterations)
            write(row, 3, result.time_spent)
            write(row, 4, self.distance_function(result.sample.distribution, result.distribution))
            has_ref = result.task.initial_guess is not None
            write(row, 5, self.tr("Yes") if has_ref else self.tr("No"))
            is_ref = result.uuid in self.__reference_map
            write(row, 6, self.tr("Yes") if is_ref else self.tr("No"))

        self.data_table.resizeColumnsToContents()

    def add_result(self, result: SSUResult):
        if self.n_results == 0 or \
            (self.page_index == self.n_pages - 1 and \
            divmod(self.n_results, self.PAGE_ROWS)[-1] != 0):
            need_update = True
        else:
            need_update = False
        self.__fitting_results.append(result)
        self.update_page_list()
        if need_update:
            self.update_page(self.page_index)

    def add_results(self, results: typing.List[SSUResult]):
        if self.n_results == 0 or \
            (self.page_index == self.n_pages - 1 and \
            divmod(self.n_results, self.PAGE_ROWS)[-1] != 0):
            need_update = True
        else:
            need_update = False
        self.__fitting_results.extend(results)
        self.update_page_list()
        if need_update:
            self.update_page(self.page_index)

    def mark_results(self, results: typing.List[SSUResult]):
        for result in results:
            self.__reference_map[result.uuid] = result

        self.update_page(self.page_index)

    def unmark_results(self, results: typing.List[SSUResult]):
        for result in results:
            if result.uuid in self.__reference_map:
                self.__reference_map.pop(result.uuid)

        self.update_page(self.page_index)

    def add_references(self, results: typing.List[SSUResult]):
        self.add_results(results)
        self.mark_results(results)

    def mark_selections(self):
        results = [self.__fitting_results[selection] for selection in self.selections]
        self.mark_results(results)

    def unmark_selections(self):
        results = [self.__fitting_results[selection] for selection in self.selections]
        self.unmark_results(results)

    def remove_results(self, indexes):
        results = []
        for i in reversed(indexes):
            res = self.__fitting_results.pop(i)
            results.append(res)
        self.unmark_results(results)
        self.update_page_list()
        self.update_page(self.page_index)

    def remove_selections(self):
        indexes = self.selections
        self.remove_results(indexes)

    def remove_all_results(self):
        res = self.remove_warning_msg.exec_()
        if res == QtWidgets.QMessageBox.Yes:
            self.__fitting_results.clear()
            self.__reference_map.clear()
            self.update_page_list()
            self.update_page(0)

    def show_distance(self):
        results = [self.__fitting_results[i] for i in self.selections]
        if results is None or len(results) == 0:
            return
        result = results[0]
        self.distance_chart.show_distance_series(
            result.get_distance_series(self.distance_name),
            ylabel=self.distance_name,
            title=result.sample.name)
        self.distance_chart.show()

    def show_chart(self):
        results = [self.__fitting_results[i] for i in self.selections]
        if results is None or len(results) == 0:
            return
        result = results[0]
        self.result_displayed.emit(result)

    def load_references(self, mark_ref=False):
        filename, _ = self.file_dialog.getOpenFileName(
            self, self.tr("Choose a file which stores the dumped SSU results"),
            None, self.tr("Dumped SSU Results (*.ssu)"))
        if filename is None or filename == "":
            return
        with open(filename, "rb") as f:
            results = pickle.load(f) # type: list[SSUResult]
            valid = True
            if isinstance(results, list):
                for result in results:
                    if not isinstance(result, SSUResult):
                        valid = False
                        break
            else:
                valid = False

            if valid:
                if self.n_results != 0 and len(results) != 0:
                    old_classes = self.__fitting_results[0].classes_φ
                    new_classes = results[0].classes_φ
                    classes_inconsistent = False
                    if len(old_classes) != len(new_classes):
                        classes_inconsistent = True
                    else:
                        classes_error = np.abs(old_classes - new_classes)
                        if not np.all(np.less_equal(classes_error, 1e-8)):
                            classes_inconsistent = True
                    if classes_inconsistent:
                        self.logger.error("The grain size classes of the SSU results in binary file are inconsistent with that in list.")
                        self.show_error(self.tr("The grain size classes of the SSU results in binary file are inconsistent with that in list."))
                        return
                self.add_results(results)
                if mark_ref:
                    self.mark_results(results)
                self.logger.info(f"There are {len(results)} SSU results that have been loaded.")
            else:
                self.logger.error("The binary file is invalid (i.e., the objects in it are not SSU results).")
                self.show_error(self.tr("The binary file is invalid (i.e., the objects in it are not SSU results)."))

    def dump_references(self):
        if self.n_results == 0:
            self.show_warning(self.tr("There is no SSU result."))
            return
        filename, _  = self.file_dialog.getSaveFileName(
            self, self.tr("Choose a filename to dump the SSU results"),
            None, self.tr("Dumped SSU Results (*.ssu)"))
        if filename is None or filename == "":
            return
        with open(filename, "wb") as f:
            pickle.dump(self.__fitting_results, f)
        self.logger.info("The SSU results have been dumped.")

    def find_similar(self, target: GrainSizeSample, ref_results: typing.List[SSUResult]):
        assert len(ref_results) != 0
        # sample_moments = logarithmic(sample.classes_φ, sample.distribution)
        # keys_to_check = ["mean", "std", "skewness", "kurtosis"]

        start_time = time.time()
        from scipy.interpolate import interp1d
        min_distance = 1e100
        min_result = None
        trans_func = interp1d(target.classes_φ, target.distribution, bounds_error=False, fill_value=0.0)
        for result in ref_results:
            trans_dist = trans_func(result.classes_φ)
            distance = self.distance_function(result.distribution, trans_dist)
            if distance < min_distance:
                min_distance = distance
                min_result = result
        # self.logger.debug(f"It took {time.time()-start_time:0.4f} s to query the reference from {len(ref_results)} results.")
        return min_result

    def query_reference(self, sample: GrainSizeSample):
        if len(self.__reference_map) == 0:
            # self.logger.debug("Try to query the reference, but there is no result that has been marked as the reference.")
            return None
        return self.find_similar(sample, self.__reference_map.values())

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("SSU Reference Viewer"))
        self.previous_button.setText(self.tr("Previous"))
        self.previous_button.setToolTip(self.tr("Click to get back to the previous page."))
        for i in range(self.page_combo_box.count()):
            self.page_combo_box.setItemText(i, self.tr("Page {0}").format(i+1))
        self.next_button.setText(self.tr("Next"))
        self.next_button.setToolTip(self.tr("Click to jump to the next page."))
        self.distance_label.setText(self.tr("Distance Function"))
        self.distance_label.setToolTip(self.tr("The function to calculate the difference (on the contrary, similarity) between two samples."))
        self.mark_action.setText(self.tr("Mark"))
        self.unmark_action.setText(self.tr("Unmark"))
        self.remove_action.setText(self.tr("Remove"))
        self.remove_all_action.setText(self.tr("Remove All"))
        self.show_chart_action.setText(self.tr("Show Chart"))
        self.show_distance_action.setText(self.tr("Show Distance Series"))
        self.update_page(self.page_index)



from uuid import uuid4

from ..model import GrainSizeSample
from ._distribution import DistributionType


class SSUTask:
    def __init__(self, sample: GrainSizeSample,
                 distribution_type: DistributionType,
                 n_components: int,
                 resolver_setting=None,
                 initial_guess=None):
        self.uuid = uuid4()
        self.sample = sample
        self.distribution_type = distribution_type
        self.n_components = n_components
        self.resolver_setting = resolver_setting
        self.initial_guess = initial_guess



__all__ = ["Reference"]

import numpy as np

from ._distribution import DISTRIBUTION_CLASS_MAP, DistributionType


class Reference:
    def __init__(self,
                 distribution_type: DistributionType,
                 params: np.ndarray):
        assert isinstance(distribution_type, DistributionType)
        assert isinstance(params, np.ndarray)
        assert params.ndim == 2
        n_params, n_components = params.shape
        assert n_params == DISTRIBUTION_CLASS_MAP[distribution_type].N_PARAMS
        self.__distribution_type = distribution_type
        self.__params = params
        self.__n_components = n_components
        self.__n_params = n_params

    @property
    def distribution_type(self) -> DistributionType:
        return self.__distribution_type

    @property
    def n_components(self) -> int:
        return self.__n_components

    @property
    def n_params(self) -> int:
        return self.__n_params

    @property
    def params(self) -> np.ndarray:
        return self.__params



import copy
import typing
from uuid import UUID, uuid4

import numpy as np
from scipy.stats import norm

from ..model import GrainSizeSample
from ._distance import get_distance_function
from ._distribution import DISTRIBUTION_CLASS_MAP, DistributionType
from ._task import SSUTask


class SSUViewModel:
    def __init__(self, classes_φ, target,
                 mixed, distributions, proportions,
                 component_prefix="C", title="", **kwargs):
        self.classes_φ = classes_φ
        self.mixed = mixed
        self.distributions = distributions
        self.proportions = proportions
        self.target=target
        self.component_prefix = component_prefix
        self.title = title
        self.kwargs = kwargs

    @property
    def n_components(self) -> int:
        return len(self.distributions)


class SSUComponentResult:
    def __init__(self, distribution: np.ndarray,
                 proportion: float,
                 moments: typing.Tuple[float, float, float, float],
                 component_args: np.ndarray):
        assert proportion is not None and np.isreal(proportion)
        # iteration may pass invalid proportion value
        # assert proportion >= 0.0 and proportion <= 1.0
        self.update(distribution, proportion, moments, component_args)

    def update(self, distribution: np.ndarray,
               proportion: float,
               moments: typing.Tuple[float, float, float, float],
               component_args: np.ndarray):
        self.__distribution = distribution
        self.__proportion = proportion
        m, v, s, k = moments
        self.__moments = dict(mean=m, std=np.sqrt(v), skewness=s, kurtosis=k)
        self.__component_args = component_args
        values_to_check = [self.__proportion]
        values_to_check.extend(self.__distribution)
        keys = ["mean", "std", "skewness", "kurtosis"]
        for key in keys:
            values_to_check.append(self.__moments[key])
        values_to_check = np.array(values_to_check)
        # if any value is nan of inf, this result is invalid
        if np.any(np.isnan(values_to_check) | np.isinf(values_to_check)):
            self.__is_valid = False
        else:
            self.__is_valid = True

    @property
    def proportion(self) -> float:
        return self.__proportion

    @property
    def distribution(self) -> np.ndarray:
        return self.__distribution

    @property
    def moments(self) -> dict:
        return self.__moments

    @property
    def component_args(self) -> np.ndarray:
        return self.__component_args

    @property
    def is_valid(self) -> bool:
        return self.__is_valid


class SSUResult:
    """
    The class to represent the fitting result of each sample.
    """
    def __init__(self, task: SSUTask,
                 func_args: typing.Iterable[float],
                 history: typing.List[np.ndarray] = None,
                 time_spent = None):
        # add uuid to manage data
        self.__uuid = uuid4()
        self.__task = task
        self.__distribution_type = task.distribution_type
        self.__n_components = task.n_components
        self.__sample = task.sample
        self.__func_args = func_args
        self.__history = [func_args] if history is None else history
        self.__components = [] # type: typing.List[SSUComponentResult]
        self.__time_spent = time_spent
        self.update(func_args)

    def update(self, func_args: typing.Iterable[float]):
        self.__func_args = func_args
        distribution_class = DISTRIBUTION_CLASS_MAP[self.distribution_type]
        classes = np.expand_dims(np.expand_dims(self.sample.classes_φ, 0), 0).repeat(1, 0).repeat(self.n_components, 1)
        if np.any(np.isnan(func_args)):
            self.__distribution = np.full_like(self.sample.distribution, fill_value=np.nan)
            self.__is_valid = False
        else:
            proportions, components, (m, v, s, k) = distribution_class.interpret(func_args, classes, self.__sample.interval_φ)
            proportions, components, (m, v, s, k) = proportions[0], components[0], (m[0], v[0], s[0], k[0])
            distribution = (proportions @ components)[0]
            self.__distribution = distribution
            if len(self.__components) == 0:
                for i in range(self.n_components):
                    component_result = SSUComponentResult(components[i], proportions[0][i], (m[i], v[i], s[i], k[i]), func_args[0, :, i])
                    self.__components.append(component_result)
            else:
                for i in range(self.n_components):
                    self.__components[i].update(components[i], proportions[0][i], (m[i], v[i], s[i], k[i]), func_args[0, :, i])
            # sort by mean φ values
            # reverse is necessary
            self.__components.sort(key=lambda component: component.moments["mean"], reverse=True)

            self.__is_valid = True
            if np.any(np.isnan(self.__distribution) | np.isinf(self.__distribution)):
                self.__is_valid = False
            for component in self.__components:
                if not component.is_valid:
                    self.__is_valid = False
                    break

    @property
    def uuid(self) -> UUID:
        return self.__uuid

    @property
    def sample(self) -> GrainSizeSample:
        return self.__sample

    @property
    def classes_μm(self) -> np.ndarray:
        return self.__sample.classes_μm

    @property
    def classes_φ(self) -> np.ndarray:
        return self.__sample.classes_φ

    @property
    def task(self) -> SSUTask:
        return self.__task

    @property
    def func_args(self) -> np.ndarray:
        return self.__func_args

    @property
    def distribution_type(self) -> DistributionType:
        return self.__distribution_type

    @property
    def n_components(self) -> int:
        return self.__n_components

    @property
    def distribution(self) -> np.ndarray:
        return self.__distribution

    @property
    def components(self) -> typing.List[SSUComponentResult]:
        return self.__components

    @property
    def is_valid(self) -> bool:
        return self.__is_valid

    @property
    def history(self):
        copy_result = copy.deepcopy(self)
        for func_args in self.__history:
            copy_result.update(func_args)
            yield copy_result

    def get_distance(self, distance: str):
        distance_func = get_distance_function(distance)
        distance = distance_func(self.distribution, self.sample.distribution)
        return distance

    def get_distance_series(self, distance: str):
        distance_series = []
        for result in self.history:
            distance_series.append(result.get_distance(distance))
        return distance_series

    @property
    def time_spent(self):
        return self.__time_spent

    @property
    def n_iterations(self):
        return len(self.__history)

    @property
    def view_model(self) -> SSUViewModel:
        distributions = [component.distribution for component in self.components]
        proportions = [comp.proportion for comp in self.components]
        return SSUViewModel(
            self.sample.classes_φ, self.sample.distribution,
            self.distribution, distributions, proportions,
            component_prefix="C", title=self.sample.name)

    @property
    def view_models(self) -> typing.Iterable[SSUViewModel]:
        for result in self.history:
            yield result.view_model



from __future__ import annotations

import typing
from uuid import UUID, uuid4

import numpy as np

from ..statistic import get_interval_φ


class ClassesNotIncrementalError(Exception):
    """Raises while the array of grain size classes is not incremental.

    It's just an ASSUMPTION for the convenience of coding.
    """
    pass


class ClassesNotMatchError(Exception):
    """Raises while the grain size classes of a new batch of sample data are not equal to the existing classes."""
    pass


class NaNError(Exception):
    """Raises while there is at least one NaN value in the array."""
    pass


class ArrayEmptyError(Exception):
    """Raises while the length of array is 0."""
    pass


class SampleNameEmptyError(Exception):
    """Raises while the name of sample is empty."""
    pass


class DistributionSumError(Exception):
    """Raises while the sum of distribution values is not equal to 1 or 100."""
    pass


class SampleNotFountError(Exception):
    """Raises while the sample is not found in this dataset."""
    pass


class GrainSizeSample:
    def __init__(self, name: str,
                 classes_μm: np.ndarray,
                 classes_φ: np.ndarray,
                 distribution: np.ndarray):
        self.__uuid = uuid4()
        self.__name = name
        self.__classes_μm = classes_μm
        self.__classes_φ = classes_φ
        self.__distribution = distribution

    @property
    def uuid(self) -> UUID:
        return self.__uuid

    @property
    def name(self) -> str:
        return self.__name

    @property
    def classes_μm(self) -> np.ndarray:
        return self.__classes_μm

    @property
    def classes_φ(self) -> np.ndarray:
        return self.__classes_φ

    @property
    def interval_φ(self) -> float:
        return get_interval_φ(self.classes_φ)

    @property
    def distribution(self) -> np.ndarray:
        return self.__distribution


class GrainSizeDataset:
    def __init__(self):
        self.__classes_μm = None # type: np.ndarray
        self.__classes_φ = None # type: np.ndarray
        self.__id_table = {} # type: dict[UUID, GrainSizeSample]
        self.__samples = [] # type: list[GrainSizeSample]

    @property
    def n_samples(self) -> int:
        return len(self.__samples)

    @property
    def n_classes(self) -> int:
        if self.__classes_μm is None:
            return 0
        else:
            return len(self.__classes_μm)

    @property
    def has_sample(self) -> bool:
        if self.n_samples == 0:
            return False
        else:
            return True

    @property
    def classes_μm(self) -> np.ndarray:
        return self.__classes_μm

    @property
    def classes_φ(self) -> np.ndarray:
        return self.__classes_φ

    @property
    def interval_φ(self) -> float:
        return get_interval_φ(self.classes_φ)

    @property
    def samples(self) -> typing.List[GrainSizeSample]:
        return self.__samples.copy()

    @property
    def distribution_matrix(self) -> np.ndarray:
        distributions = [sample.distribution for sample in self.samples]
        matrix = np.array(distributions)
        return matrix

    @staticmethod
    def is_incremental(nums: np.ndarray) -> bool:
        """Returns `True` while the array is incremental.
        This method is used to validate the array of grain size classes."""
        # for i in range(1, len(nums)):
        #     if nums[i] <= nums[i-1]:
        #         return False
        # return True
        return np.all(np.less(nums[:-1], nums[1:]))

    @staticmethod
    def validate_classes_μm(classes: np.ndarray):
        assert classes is not None
        assert type(classes) == np.ndarray
        assert classes.dtype == np.float64
        if len(classes) == 0:
            raise ArrayEmptyError("The array of grain size classes is empty.")
        if np.any(np.isnan(classes)):
            raise NaNError("There is at least one NaN value in grain size classes.")
        # It may raises when the users select a wrong data file.
        if not GrainSizeDataset.is_incremental(classes):
            raise ClassesNotIncrementalError("The array of grain size classes is not incremental.")

    @staticmethod
    def validate_sample_name(sample_name: str):
        assert sample_name is not None
        assert type(sample_name) == str
        if sample_name == "":
            raise SampleNameEmptyError("Sample name is empty.")

    @staticmethod
    def validate_distribution(distribution: np.ndarray):
        assert distribution is not None
        assert type(distribution) == np.ndarray
        assert distribution.dtype == np.float64
        if len(distribution) == 0:
            raise ArrayEmptyError("The array of distribution of this sample is empty.")
        if np.any(np.isnan(distribution)):
            raise NaNError("There is at least one NaN value in the distribution.")
        # Check if the sum is close to 1 or 100 (some users may use percentage).
        s = np.sum(distribution)
        if (s > 0.95 and s < 1.05):
            pass
        elif (s > 95 and s < 105):
            # If the sum is close to 100, make it close to 1.
            np.true_divide(distribution, 100.0, out=distribution)
        else:
            raise DistributionSumError("Distribution of this sample not sum equal to 1 or 100.")

    @staticmethod
    def are_classes_match(classes_a: np.ndarray, classes_b: np.ndarray):
        if len(classes_a) != len(classes_b):
            return False
        equal_res = np.equal(classes_a, classes_b)
        if not np.all(equal_res):
            return False
        return True

    def add_batch(self, classes_μm: np.ndarray,
                  names: typing.List[str],
                  distributions: typing.List[np.ndarray],
                  need_validation: bool = True):
        self.validate_classes_μm(classes_μm)
        # If there is any sample already, it's necessary to check the consistency of classes.
        if self.has_sample:
            if not self.are_classes_match(self.classes_μm, classes_μm):
                raise ClassesNotMatchError(self.classes_μm, classes_μm)
        assert len(names) == len(distributions)

        self.__classes_μm = classes_μm
        self.__classes_φ = -np.log2(classes_μm/1000.0)
        # use temp to implement roll-back
        temp_table = {}
        temp_list = []
        for name, distribution in zip(names, distributions):
            assert len(distribution) == len(classes_μm)
            if need_validation:
                self.validate_sample_name(name)
                self.validate_distribution(distribution)
            sample = GrainSizeSample(name, self.classes_μm, self.classes_φ, distribution)
            temp_table.update({sample.uuid: sample})
            temp_list.append(sample)
        # if no exception raised, the codes below will be executed
        self.__id_table.update(temp_table)
        self.__samples.extend(temp_list)

    def combine(self, another_dataset: GrainSizeDataset):
        if self.are_classes_match(self.classes_μm, another_dataset.classes_μm):
            self.__id_table.update(another_dataset.__id_table)
            self.__samples.extend(another_dataset.__samples)
        else:
            raise ClassesNotMatchError(self.classes_μm, another_dataset.classes_μm)

    def get_sample_by_id(self, uuid: UUID):
        if uuid in self.__id_table:
            return self.__id_table[uuid]
        else:
            raise SampleNotFountError("There is no sample with this id.", uuid)



import typing

import matplotlib.pyplot as plt
import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets

from ..model import GrainSizeSample
from ..statistic import convert_φ_to_μm, get_cumulative_frequency
from .BaseChart import BaseChart
from .config_matplotlib import normal_color


class CumulativeCurveChart(BaseChart):
    def __init__(self, parent=None, figsize=(4, 3)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("Cumulative Curve Chart"))
        self.axes = self.figure.subplots()

        self.scale_menu = QtWidgets.QMenu(self.tr("Scale")) # type: QtWidgets.QMenu
        self.menu.insertMenu(self.save_figure_action, self.scale_menu)
        self.scale_group = QtGui.QActionGroup(self.scale_menu)
        self.scale_group.setExclusive(True)
        self.scale_actions = [] # type: list[QtGui.QAction]
        for key, name in self.supported_scales:
            scale_action = self.scale_group.addAction(name) # type: QtGui.QAction
            scale_action.setCheckable(True)
            scale_action.triggered.connect(self.update_chart)
            self.scale_menu.addAction(scale_action)
            self.scale_actions.append(scale_action)
        self.scale_actions[0].setChecked(True)

        self.__last_samples = []

    @property
    def supported_scales(self) -> typing.List[typing.Tuple[str, str]]:
        scales = [("log-linear", self.tr("Log-linear")),
                  ("log", self.tr("Log")),
                  ("phi", self.tr("Phi")),
                  ("linear", self.tr("Linear"))]
        return scales

    @property
    def scale(self) -> str:
        for i, scale_action in enumerate(self.scale_actions):
            if scale_action.isChecked():
                key, name = self.supported_scales[i]
                return key

    @property
    def transfer(self) -> typing.Callable:
        if self.scale == "log-linear":
            return lambda classes_φ: convert_φ_to_μm(classes_φ)
        elif self.scale == "log":
            return lambda classes_φ: np.log(convert_φ_to_μm(classes_φ))
        elif self.scale == "phi":
            return lambda classes_φ: classes_φ
        elif self.scale == "linear":
            return lambda classes_φ: convert_φ_to_μm(classes_φ)

    @property
    def xlabel(self) -> str:
        if self.scale == "log-linear":
            return "Grain size [μm]"
        elif self.scale == "log":
            return "Ln(grain size in μm)"
        elif self.scale == "phi":
            return "Grain size [φ]"
        elif self.scale == "linear":
            return "Grain size [μm]"

    @property
    def ylabel(self) -> str:
        return "Frequency"

    @property
    def xlog(self) -> bool:
        if self.scale == "log-linear":
            return True
        else:
            return False

    def update_chart(self):
        self.figure.clear()
        self.axes = self.figure.subplots()
        self.show_samples(self.__last_samples, append=False)

    def show_samples(self, samples: typing.Iterable[GrainSizeSample], append=False):
        if len(samples) == 0:
            return
        append = append and len(self.__last_samples) != 0
        if not append:
            self.axes.clear()
            self.__last_samples = []
        for i, sample in enumerate(samples):
            self.__last_samples.append(sample)
            if i == 0:
                x = self.transfer(sample.classes_φ)
                if not append:
                    if self.xlog:
                        self.axes.set_xscale("log")
                    self.axes.set_title("Cumulative curves of samples")
                    self.axes.set_xlabel(self.xlabel)
                    self.axes.set_ylabel(self.ylabel)
                    self.axes.set_xlim(x[0], x[-1])
                    self.axes.set_ylim(0.0, 1.0)
            cumulative_frequency = get_cumulative_frequency(sample.distribution)
            self.axes.plot(x, cumulative_frequency, c=normal_color())
        self.figure.tight_layout()
        self.canvas.draw()

    def retranslate(self):
        super().retranslate()
        self.setWindowTitle(self.tr("Cumulative Curve Chart"))
        self.scale_menu.setTitle(self.tr("Scale"))



import typing

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FFMpegWriter, FuncAnimation, ImageMagickWriter
from PySide6 import QtCore, QtGui, QtWidgets

from ..emma import EMMAResult
from ..statistic import convert_φ_to_μm
from .BaseChart import BaseChart
from .config_matplotlib import normal_color


class EMMAResultChart(BaseChart):
    N_DISPLAY_SAMPLES = 200
    def __init__(self, parent=None, figsize=(4, 6)):
        super().__init__(parent=parent, figsize=figsize)
        # self.axes = self.figure.subplots()
        self.scale_menu = QtWidgets.QMenu(self.tr("Scale")) # type: QtWidgets.QMenu
        self.menu.insertMenu(self.save_figure_action, self.scale_menu)
        self.scale_group = QtGui.QActionGroup(self.scale_menu)
        self.scale_group.setExclusive(True)
        self.scale_actions = [] # type: list[QtGui.QAction]
        for key, name in self.supported_scales:
            scale_action = self.scale_group.addAction(name) # type: QtGui.QAction
            scale_action.setCheckable(True)
            scale_action.triggered.connect(self.update_chart)
            self.scale_menu.addAction(scale_action)
            self.scale_actions.append(scale_action)
        self.scale_actions[0].setChecked(True)

        self.distance_menu = QtWidgets.QMenu(self.tr("Distance Function")) # type: QtWidgets.QMenu
        self.menu.insertMenu(self.save_figure_action, self.distance_menu)
        self.distance_group = QtGui.QActionGroup(self.distance_menu)
        self.distance_group.setExclusive(True)
        self.distance_actions = [] # type: list[QtGui.QAction]
        for key, name in self.supported_distances:
            distance_action = self.distance_group.addAction(name) # type: QtGui.QAction
            distance_action.setCheckable(True)
            distance_action.triggered.connect(self.update_chart)
            self.distance_menu.addAction(distance_action)
            self.distance_actions.append(distance_action)
        self.distance_actions[5].setChecked(True)

        self.animated_action = QtGui.QAction(self.tr("Animated")) # type: QtGui.QAction
        self.animated_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.save_figure_action, self.animated_action)
        self.animated_action.setCheckable(True)
        self.animated_action.setChecked(False)

        self.interval_menu = QtWidgets.QMenu(self.tr("Animation Interval")) # type: QtWidgets.QMenu
        self.menu.insertMenu(self.save_figure_action, self.interval_menu)
        self.interval_group = QtGui.QActionGroup(self.interval_menu)
        self.interval_group.setExclusive(True)
        self.interval_actions = [] # type: list[QtGui.QAction]
        for interval, name in self.supported_intervals:
            interval_action = self.interval_group.addAction(name)
            interval_action.setCheckable(True)
            interval_action.triggered.connect(self.update_chart)
            self.interval_menu.addAction(interval_action)
            self.interval_actions.append(interval_action)
        self.interval_actions[3].setChecked(True)

        self.repeat_action = QtGui.QAction(self.tr("Repeat Animation")) # type: QtGui.QAction
        self.repeat_action.triggered.connect(self.update_chart)
        self.menu.insertAction(self.save_figure_action, self.repeat_action)
        self.repeat_action.setCheckable(True)
        self.repeat_action.setChecked(False)

        self.save_animation_action = QtGui.QAction(self.tr("Save Animation")) # type: QtGui.QAction
        self.menu.addAction(self.save_animation_action)
        self.save_animation_action.triggered.connect(self.save_animation)

        self.__animation = None
        self.__last_result = None

        self.normal_msg = QtWidgets.QMessageBox(parent=self)
        self.file_dialog = QtWidgets.QFileDialog(parent=self)

    @property
    def supported_scales(self) -> typing.List[typing.Tuple[str, str]]:
        scales = [("log-linear", self.tr("Log-linear")),
                  ("log", self.tr("Log")),
                  ("phi", self.tr("Phi")),
                  ("linear", self.tr("Linear"))]
        return scales

    @property
    def supported_distances(self) -> typing.List[typing.Tuple[str, str]]:
        distances = [
            ("1-norm", self.tr("1 Norm")),
            ("2-norm", self.tr("2 Norm")),
            ("3-norm", self.tr("3 Norm")),
            ("4-norm", self.tr("4 Norm")),
            ("MSE", self.tr("Mean Squared Error")),
            ("log10MSE", self.tr("Logarithmic Mean Squared Error")),
            ("cosine", self.tr("Cosine Error")),
            ("angular", self.tr("Angular Error"))]
        return distances

    @property
    def supported_intervals(self) -> typing.List[typing.Tuple[int, str]]:
        intervals = [(5, self.tr("5 Milliseconds")),
                     (10, self.tr("10 Milliseconds")),
                     (20, self.tr("20 Milliseconds")),
                     (30, self.tr("30 Milliseconds")),
                     (60, self.tr("60 Milliseconds"))]
        return intervals

    @property
    def scale(self) -> str:
        for i, scale_action in enumerate(self.scale_actions):
            if scale_action.isChecked():
                key, name = self.supported_scales[i]
                return key

    @property
    def distance(self) -> str:
        for i, distance_action in enumerate(self.distance_actions):
            if distance_action.isChecked():
                key, name = self.supported_distances[i]
                return key

    @property
    def animated(self) -> bool:
        return self.animated_action.isChecked()

    @property
    def animation_interval(self) -> int:
        for i, interval_action in enumerate(self.interval_actions):
            if interval_action.isChecked():
                interval, name = self.supported_intervals[i]
                return interval

    @property
    def repeat_animation(self) -> bool:
        return self.repeat_action.isChecked()

    @property
    def transfer(self) -> typing.Callable:
        if self.scale == "log-linear":
            return lambda classes_φ: convert_φ_to_μm(classes_φ)
        elif self.scale == "log":
            return lambda classes_φ: np.log(convert_φ_to_μm(classes_φ))
        elif self.scale == "phi":
            return lambda classes_φ: classes_φ
        elif self.scale == "linear":
            return lambda classes_φ: convert_φ_to_μm(classes_φ)

    @property
    def xlabel(self) -> str:
        if self.scale == "log-linear":
            return "Grain size [μm]"
        elif self.scale == "log":
            return "Ln(grain size in μm)"
        elif self.scale == "phi":
            return "Grain size [φ]"
        elif self.scale == "linear":
            return "Grain size [μm]"

    @property
    def ylabel(self) -> str:
        return "Frequency"

    @property
    def xlog(self) -> bool:
        if self.scale == "log-linear":
            return True
        else:
            return False

    def show_chart(self, result: EMMAResult):
        classes = self.transfer(result.dataset.classes_φ)
        sample_indexes = np.linspace(1, result.n_samples, result.n_samples)
        iteration_indexes = np.linspace(1, result.n_iterations, result.n_iterations)
        interval = max(1, result.n_samples // self.N_DISPLAY_SAMPLES)
        GSDs = result.dataset.distribution_matrix

        GSDs_axes = self.figure.add_subplot(2, 2, 1)
        for sample in result.dataset.samples[::interval]:
            GSDs_axes.plot(classes, sample.distribution, c=normal_color(), alpha=0.2)
        if self.xlog:
            GSDs_axes.set_xscale("log")
        GSDs_axes.set_xlim(classes[0], classes[-1])
        GSDs_axes.set_ylim(0.0, round(np.max(GSDs)*1.2, 2))
        GSDs_axes.set_xlabel(self.xlabel)
        GSDs_axes.set_ylabel(self.ylabel)
        GSDs_axes.set_title("GSDs")

        distance_axes = self.figure.add_subplot(2, 2, 2)
        distance_axes.plot(iteration_indexes, result.get_distance_series(self.distance), c=normal_color())
        distance_axes.set_xlim(iteration_indexes[0], iteration_indexes[-1])
        distance_axes.set_xlabel("Iteration")
        distance_axes.set_ylabel("Distance")
        distance_axes.set_title("Distance variation")

        # get the mode size of each end-members
        modes = [(i, result.dataset.classes_μm[np.unravel_index(np.argmax(result.end_members[i]), result.end_members[i].shape)]) for i in range(result.n_members)]
        # sort them by mode size
        modes.sort(key=lambda x: x[1])
        end_member_axes = self.figure.add_subplot(2, 2, 3)
        if self.xlog:
            end_member_axes.set_xscale("log")
        for i_em, (i, _) in enumerate(modes):
            end_member_axes.plot(classes, result.end_members[i], c=plt.get_cmap()(i_em), label=f"EM{i_em+1}", zorder=10+i_em)
        end_member_axes.set_xlim(classes[0], classes[-1])
        end_member_axes.set_ylim(0.0, round(np.max(result.end_members)*1.2, 2))
        end_member_axes.set_xlabel(self.xlabel)
        end_member_axes.set_ylabel(self.ylabel)
        end_member_axes.set_title("End members")

        proportion_axes = self.figure.add_subplot(2, 2, 4)
        bottom = np.zeros(result.n_samples)
        for i_em, (i, _) in enumerate(modes):
            proportion_axes.bar(
                sample_indexes[::interval],
                result.proportions[:, i][::interval],
                bottom=bottom[::interval], width=interval, color=plt.get_cmap()(i_em))
            bottom += result.proportions[:, i]
        proportion_axes.set_xlim(sample_indexes[0], sample_indexes[-1])
        proportion_axes.set_ylim(0.0, 1.0)
        proportion_axes.set_xlabel("Sample index")
        proportion_axes.set_ylabel("Proportion")
        proportion_axes.set_title("Proportions")
        self.figure.tight_layout()
        self.canvas.draw()

    def show_animation(self, result: EMMAResult):
        classes = self.transfer(result.dataset.classes_φ)
        sample_indexes = np.linspace(1, result.n_samples, result.n_samples)
        iteration_indexes = np.linspace(1, result.n_iterations, result.n_iterations)
        distance_series = result.get_distance_series(self.distance)
        min_distance, max_distance = np.min(distance_series), np.max(distance_series)
        interval = max(1, result.n_samples // self.N_DISPLAY_SAMPLES)
        GSDs = result.dataset.distribution_matrix

        GSDs_axes = self.figure.add_subplot(2, 2, 1)
        for sample in result.dataset.samples[::interval]:
            GSDs_axes.plot(classes, sample.distribution, c=normal_color(), alpha=0.2)
        if self.xlog:
            GSDs_axes.set_xscale("log")
        GSDs_axes.set_xlim(classes[0], classes[-1])
        GSDs_axes.set_ylim(0.0, round(np.max(GSDs)*1.2, 2))
        GSDs_axes.set_xlabel(self.xlabel)
        GSDs_axes.set_ylabel(self.ylabel)
        GSDs_axes.set_title("GSDs")

        distance_axes = self.figure.add_subplot(2, 2, 2)
        distance_axes.plot(iteration_indexes, result.get_distance_series(self.distance), c=normal_color())
        distance_axes.set_xlim(iteration_indexes[0], iteration_indexes[-1])
        distance_axes.set_xlabel("Iteration")
        distance_axes.set_ylabel("Distance")
        distance_axes.set_title("Distance variation")

        # get the mode size of each end-members
        modes = [(i, result.dataset.classes_μm[np.unravel_index(np.argmax(result.end_members[i]), result.end_members[i].shape)]) for i in range(result.n_members)]
        # sort them by mode size
        modes.sort(key=lambda x: x[1])
        end_member_axes = self.figure.add_subplot(2, 2, 3)
        if self.xlog:
            end_member_axes.set_xscale("log")
        end_member_axes.set_xlim(classes[0], classes[-1])
        end_member_axes.set_ylim(0.0, round(np.max(result.end_members)*1.2, 2))
        end_member_axes.set_xlabel(self.xlabel)
        end_member_axes.set_ylabel(self.ylabel)
        end_member_axes.set_title("End members")

        proportion_axes = self.figure.add_subplot(2, 2, 4)
        proportion_axes.set_xlim(sample_indexes[0], sample_indexes[-1])
        proportion_axes.set_ylim(0.0, 1.0)
        proportion_axes.set_xlabel("Sample index")
        proportion_axes.set_ylabel("Proportion")
        proportion_axes.set_title("Proportions")

        # self.figure.tight_layout()
        # self.canvas.draw()
        def init():
            self.iteration_line = distance_axes.plot([1, 1], [min_distance, max_distance], c=normal_color())[0]
            self.end_member_curves = []
            for i_em, (index, _) in enumerate(modes):
                end_member_curve = end_member_axes.plot(classes, result.end_members[index], c=plt.get_cmap()(i_em), label=f"EM{i_em+1}")[0]
                self.end_member_curves.append(end_member_curve)
            bottom = np.zeros(result.n_samples)
            self.proportion_bars = []
            self.patches = []
            for i_em, (index, _) in enumerate(modes):
                bar = proportion_axes.bar(sample_indexes[::interval], result.proportions[:, index][::interval], bottom=bottom[::interval], width=interval, color=plt.get_cmap()(i_em))
                self.proportion_bars.append(bar)
                self.patches.extend(bar.patches)
                bottom += result.proportions[:, index]
            return self.iteration_line, *(self.end_member_curves + self.patches)

        def animate(args: typing.Tuple[int, EMMAResult]):
            iteration, current = args
            self.iteration_line.set_xdata([iteration, iteration])
            for i_em, (index, _) in enumerate(modes):
                self.end_member_curves[i_em].set_ydata(current.end_members[index])
            bottom = np.zeros(current.n_samples)
            for i_em, (index, _) in enumerate(modes):
                for rect, height, y in zip(self.proportion_bars[i_em].patches, current.proportions[:, index][::interval], bottom[::interval]):
                    rect.set_height(height)
                    rect.set_y(y)
                bottom += current.proportions[:, index]
            return self.iteration_line, *(self.end_member_curves + self.patches)

        self.__animation = FuncAnimation(
            self.figure, animate, init_func=init,
            frames=enumerate(result.history), interval=self.animation_interval,
            blit=True, repeat=self.repeat_animation,
            repeat_delay=3.0, save_count=result.n_iterations)

    def show_result(self, result: EMMAResult):
        self.__last_result = result
        self.figure.clear()
        if self.__animation is not None:
            self.__animation._stop()
            self.__animation = None
        if self.animated:
            self.show_animation(result)
        else:
            self.show_chart(result)

    def update_chart(self):
        if self.__last_result is not None:
            self.show_result(self.__last_result)

    def save_animation(self):
        if self.__last_result is not None:
            filename, format_str  = self.file_dialog.getSaveFileName(
                self, self.tr("Choose a filename to save the animation of this EMMA result"),
                None, "MPEG-4 Video File (*.mp4);;Graphics Interchange Format (*.gif)")
            if filename is None or filename == "":
                return
            progress = QtWidgets.QProgressDialog(self)
            progress.setWindowTitle("QGrain")
            progress.setRange(0, 100)
            progress.setLabelText(self.tr("Saving Animation [{0} Frames]").format(self.__last_result.n_iterations))
            canceled = False
            def save_callback(i, n):
                if progress.wasCanceled():
                    nonlocal canceled
                    canceled = True
                    raise StopIteration()
                progress.setValue((i+1)/n*100)
                QtCore.QCoreApplication.processEvents()

            self.animated_action.setChecked(True)
            self.update_chart()
            # plt.rcParams["savefig.dpi"] = 120.0
            if "*.gif" in format_str:
                if not ImageMagickWriter.isAvailable():
                    self.normal_msg.setWindowTitle(self.tr("Error"))
                    self.normal_msg.setText(self.tr("ImageMagick is not installed, please download and install it from its offical website (https://imagemagick.org/index.php)."))
                    self.normal_msg.exec_()
                else:
                    self.__animation.save(filename, writer="imagemagick", fps=30, progress_callback=save_callback)
            elif "*.mp4" in format_str:
                if not FFMpegWriter.isAvailable():
                    self.normal_msg.setWindowTitle(self.tr("Error"))
                    self.normal_msg.setText(self.tr("FFMpeg is not installed, please download and install it from its offical website (https://ffmpeg.org/)."))
                    self.normal_msg.exec_()
                else:
                    self.__animation.save(filename, writer="ffmpeg", fps=30, progress_callback=save_callback)
            # plt.rcParams["savefig.dpi"] = 300.0
            if not canceled:
                progress.setValue(100)

    def retranslate(self):
        super().retranslate()
        self.scale_menu.setTitle(self.tr("Scale"))
        for action, (key, name) in zip(self.scale_actions, self.supported_scales):
            action.setText(name)
        self.distance_menu.setTitle(self.tr("Distance Function"))
        for action, (key, name) in zip(self.distance_actions, self.supported_distances):
            action.setText(name)
        self.animated_action.setText(self.tr("Animated"))
        self.interval_menu.setTitle(self.tr("Animation Interval"))
        for action, (interval, name) in zip(self.interval_actions, self.supported_intervals):
            action.setText(name)
        self.repeat_action.setText(self.tr("Repeat Animation"))
        self.save_animation_action.setText(self.tr("Save Animation"))



__all__ = ["LoadDatasetDialog"]

import logging
import os
import typing

import openpyxl
import xlrd
from PySide6 import QtCore, QtWidgets

from ..io import *
from ..model import GrainSizeDataset


class LoadDatasetDialog(QtWidgets.QDialog):
    logger = logging.getLogger("QGrain.LoadDatasetDialog")
    dataset_loaded = QtCore.Signal(GrainSizeDataset)
    def __init__(self, parent=None):
        super().__init__(parent=parent, f=QtCore.Qt.Window)
        self.setWindowTitle(self.tr("Dataset Loader"))
        self.initialize_ui()
        self.normal_msg = QtWidgets.QMessageBox(self)
        self.file_dialog = QtWidgets.QFileDialog(parent=self)
        self.filename = None # type: str
        self.workbook = None # type: typing.Union[openpyxl.Workbook, xlrd.Book]
        self.dataset = None # type: GrainSizeDataset

    def initialize_ui(self):
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)

        self.filename_display = QtWidgets.QLabel(self.tr("Filename Unknown"))
        self.main_layout.addWidget(self.filename_display, 0, 0)
        self.select_button = QtWidgets.QPushButton(self.tr("Select"))
        self.select_button.clicked.connect(self.on_select_clicked)
        self.main_layout.addWidget(self.select_button, 0, 1)
        self.sheet_label = QtWidgets.QLabel(self.tr("Sheet Name"))
        self.sheet_combo_box = QtWidgets.QComboBox()
        self.sheet_combo_box.addItem(self.tr("Empty"))
        self.main_layout.addWidget(self.sheet_label, 1, 0)
        self.main_layout.addWidget(self.sheet_combo_box, 1, 1)

        self.classes_row_label = QtWidgets.QLabel(self.tr("Row With Grain Size Classes"))
        self.classes_row_input = QtWidgets.QSpinBox()
        self.classes_row_input.setRange(1, 999)
        self.main_layout.addWidget(self.classes_row_label, 2, 0)
        self.main_layout.addWidget(self.classes_row_input, 2, 1)
        self.sample_names_column_label = QtWidgets.QLabel(self.tr("Column With Sample Names"))
        self.sample_names_column_input = QtWidgets.QSpinBox()
        self.sample_names_column_input.setRange(1, 999)
        self.main_layout.addWidget(self.sample_names_column_label, 3, 0)
        self.main_layout.addWidget(self.sample_names_column_input, 3, 1)
        self.distribution_start_row_label = QtWidgets.QLabel(self.tr("Distribution Start Row"))
        self.distribution_start_row_input = QtWidgets.QSpinBox()
        self.distribution_start_row_input.setRange(2, 999999)
        self.main_layout.addWidget(self.distribution_start_row_label, 4, 0)
        self.main_layout.addWidget(self.distribution_start_row_input, 4, 1)
        self.distribution_start_column_label = QtWidgets.QLabel(self.tr("Distribution Start Column"))
        self.distribution_start_column_input = QtWidgets.QSpinBox()
        self.distribution_start_column_input.setRange(2, 999999)
        self.main_layout.addWidget(self.distribution_start_column_label, 5, 0)
        self.main_layout.addWidget(self.distribution_start_column_input, 5, 1)

        self.try_load_button = QtWidgets.QPushButton(self.tr("Try Load"))
        self.try_load_button.clicked.connect(self.on_try_load_clicked)
        self.try_load_button.setEnabled(False)
        self.main_layout.addWidget(self.try_load_button, 6, 0, 1, 2)

    @property
    def sheet_index(self) -> int:
        sheet_index = self.sheet_combo_box.currentIndex()
        return sheet_index
    @property
    def sheet_name(self) -> str:
        sheet_name = self.sheet_combo_box.currentText()
        return sheet_name

    @property
    def data_layout(self):
        classes_row = self.classes_row_input.value() - 1
        sample_names_column = self.sample_names_column_input.value() - 1
        distribution_start_row = self.distribution_start_row_input.value() - 1
        distribution_start_column = self.distribution_start_column_input.value() - 1
        try:
            layout = DataLayoutSetting(
                classes_row=classes_row,
                sample_names_column=sample_names_column,
                distribution_start_row=distribution_start_row,
                distribution_start_column=distribution_start_column)
            return layout
        except DataLayoutError as e:
            self.logger.exception("The current layout setting is invalid.", stack_info=True)
            self.show_error(self.tr("The current layout setting is invalid."))
            return None

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def on_select_clicked(self):
        filename, _ = self.file_dialog.getOpenFileName(
            self, self.tr("Select a file"), None,
            "Excel (*.xlsx);;97-2003 Excel (*.xls);;CSV (*.csv)")
        if filename is None or filename == "":
            return
        self.filename = filename
        file_type = get_type_by_name(filename)
        if file_type == ReadFileType.CSV:
            sheet_names = [os.path.basename(filename)]
        elif file_type == ReadFileType.XLS:
            self.workbook = xlrd.open_workbook(filename)
            sheet_names = self.workbook.sheet_names()
        elif file_type == ReadFileType.XLSX:
            self.workbook = openpyxl.load_workbook(filename, read_only=True, data_only=True)
            sheet_names = self.workbook.sheetnames
        else:
            raise NotImplementedError(file_type)

        self.filename_display.setText(os.path.basename(filename))
        self.sheet_combo_box.clear()
        self.sheet_combo_box.addItems(sheet_names)
        self.try_load_button.setEnabled(True)

    def on_try_load_clicked(self):
        try:
            progress_dialog = QtWidgets.QProgressDialog(
                self.tr("Loading grain size distributions..."), self.tr("Cancel"),
                0, 100, self)
            progress_dialog.setWindowTitle("QGrain")
            progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
            def callback(progress: float):
                if progress_dialog.wasCanceled():
                    raise StopIteration()
                progress_dialog.setValue(int(progress*100))
                QtCore.QCoreApplication.processEvents()
            result = load_dataset(self.filename, self.sheet_index, self.data_layout, progress_callback=callback)
            progress_dialog.setValue(100)
            if result is None:
                self.logger.exception("Can not load the grain size dataset from this file. Please check the logs for more details.", stack_info=True)
                self.show_error(self.tr("Can not load the grain size dataset from this file. Please check the logs for more details."))
            else:
                self.dataset_loaded.emit(result)
                self.logger.info("Good job! The grain size dataset has been loaded from this file, and has been emitted to other widgets.")
                self.hide()
        except StopIteration as e:
            self.logger.info("Loading task was canceled.")
            progress_dialog.close()
        except Exception as e:
            self.logger.exception("An unknown exception was raised. Please check the logs for more details.", stack_info=True)
            self.show_error(self.tr("An unknown exception was raised. Please check the logs for more details."))

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("Dataset Loader"))
        if self.filename is None:
            self.filename_display.setText(self.tr("Filename Unspecified"))
            self.sheet_combo_box.setItemText(0, self.tr("Empty"))
        self.select_button.setText(self.tr("Select"))
        self.sheet_label.setText(self.tr("Sheet Name"))
        self.classes_row_label.setText(self.tr("Row With Grain Size Classes"))
        self.sample_names_column_label.setText(self.tr("Column With Sample Names"))
        self.distribution_start_row_label.setText(self.tr("Distribution Start Row"))
        self.distribution_start_column_label.setText(self.tr("Distribution Start Column"))
        self.try_load_button.setText(self.tr("Try Load"))
